@@ -1,6 +1,7 @@
 import time
 import numpy as np
 import numba
+from numba import jit,njit, prange, types
 import warnings
 from kernelml import KernelML
 import sys
@@ -46,9 +47,12 @@ class DensityFactorization():
             combo_len = len(self.dim_combos)
             
             count=0
+            _y_ = None
             for i,j in self.dim_combos:
                 _X_ = X[:,[i,j]]
                 bins3 = [self.dim_bins[i],self.dim_bins[j]]
+                if y is not None and self.multi_dim==True:
+                    _y_ = y[:,[i,j]]
                 data,_ = hdre_histogram(_X_,y,agg_func='count',cost=None,bins=bins3)
                 data=data/np.sum(data)
                 new_kde = np.fft.fftshift(np.real(np.fft.ifftn(np.fft.fftn(data)*self.fftkernel)))
@@ -56,7 +60,7 @@ class DensityFactorization():
                 count+=1
 
     def optimize(self,X,y=None,cost=None, number_of_random_simulations=1000, number_of_realizations=10, agg_func='count',verbose=True,
-        dim_tuples=None
+        multi_dim=False,
         ):
         
         self.simulations = number_of_random_simulations
@@ -66,6 +70,7 @@ class DensityFactorization():
             raise ArgumentError("The number of bins per dimensions must be an odd integer")
         
         self.num_dim = X.shape[1]
+        self.multi_dim=multi_dim
         
         maxs = np.max(X,axis=0)
         mins = np.min(X,axis=0)
@@ -78,20 +83,25 @@ class DensityFactorization():
         
         self.dim_bins = [np.linspace(m0,m1,self.bins_per_dim) for m0,m1 in zip(mins,maxs)]
         
+        self.dim_combos = [(i,j) for i in range(X.shape[1]) for j in range(X.shape[1]) if j>i]
         #custom dimensional pair TODO
-        if dim_tuples is None:
-            self.dim_combos = [(i,j) for i in range(X.shape[1]) for j in range(X.shape[1]) if j>i]
-        else:
-            self.dim_combos=dim_tuples
+        #if dim_tuples is None:
+        #    self.dim_combos = [(i,j) for i in range(X.shape[1]) for j in range(X.shape[1]) if j>i]
+        #else:
+        #    self.dim_combos=dim_tuples
         
         combo_len = len(self.dim_combos)
         
+        _y_ = None
         if self.kde_target is None:
-            self.kde_target = np.zeros((combo_len,self.bins_per_dim,self.bins_per_dim))
-            self.bin_combos = np.zeros((combo_len,self.bins_per_dim,2))
+            self.kde_target = np.zeros((combo_len,self.bins_per_dim,self.bins_per_dim),dtype=np.float64)
+            self.bin_combos = np.zeros((combo_len,self.bins_per_dim,2),dtype=np.float64)
             count=0
             for i,j in self.dim_combos:
                 _X_ = X[:,[i,j]]
+                if y is not None and self.multi_dim==True:
+                    _y_ = y[:,[i,j]]
+
                 bins3 = [self.dim_bins[i],self.dim_bins[j]]
     #             bins3 = [np.concatenate([[-np.inf],_bins_,[np.inf]]) for _bins_ in bins3]
                 self.bin_combos[count] = np.column_stack(bins3)
@@ -114,7 +124,7 @@ class DensityFactorization():
 
         param_to_dim = np.arange(0,self.num_dim*self.num_clusters)%self.num_dim
         param_to_dim = param_to_dim.astype(np.float)
-        self.dim_combos = np.array(self.dim_combos).astype(np.int)
+        self.dim_combos = np.array(self.dim_combos).astype(np.int64)
         args = [self.dim_combos,self.kde_target,self.bin_combos,
                     min_lim,max_lim,param_to_dim,self.num_dim,widths]
         self.args = args
@@ -289,11 +299,15 @@ class HierarchicalDensityFactorization():
                  bins_per_dimension=61,
                  smoothing_parameter=1.0,
                  min_leaf_samples=1,
+                 agg_func='count',
+                 multi_dim=False,
                  alpha = 0.1,
                  verbose=0):
         
         self.alpha = alpha
         self.min_leaf_samples = min_leaf_samples
+        self.agg_func = agg_func
+        self.multi_dim = multi_dim
         self.assignments = []
         self.centroids = []
         self.deviations = []
@@ -361,7 +375,7 @@ class HierarchicalDensityFactorization():
                 
         return mask
     
-    def factorize(self,X,M):
+    def factorize(self,X,M,Y=None):
 
         
         """
@@ -377,12 +391,16 @@ class HierarchicalDensityFactorization():
         self.model.kmldata=None
         
 #         model.prune_clusters(y,pad=2.0,limit=1)
+        y=X
+        if Y is not None:
+            y=Y
+            
         
-        self.model.optimize(X=X,y=X,  number_of_random_simulations=self.number_of_random_simulations,
+        self.model.optimize(X=X,y=y,  number_of_random_simulations=self.number_of_random_simulations,
                    number_of_realizations = realizations,
-                   agg_func='count',
+                   agg_func=self.agg_func,
                    verbose=self.verbose,
-                   dim_tuples=self.dim_tuples
+                   multi_dim=self.multi_dim
                      )
         
 #         if self.count==0:
@@ -434,11 +452,13 @@ class HierarchicalDensityFactorization():
                 centers.append(C[i:i+1])
                 
                 
-        self.assignments.append(assigns)
-        self.centroids.append(np.vstack(centers))
-        self.deviations.append(S)
+        if len(centers)>0:
+            centers = np.vstack(centers)
+            self.assignments.append(assigns)
+            self.centroids.append(centers)
+            self.deviations.append(S)
             
-    def optimize(self,X,
+    def optimize(self,X,Y=None,
                  maxiter=20,
                  stop_thrshld=0.01,
                  realizations=10,
@@ -453,7 +473,7 @@ class HierarchicalDensityFactorization():
         M = np.ones(X.shape[0],dtype=bool)
         LAST = M
         for _ in range(maxiter):
-            self.factorize(X,M)
+            self.factorize(X,M,Y)
             E = np.hstack(sum(self.assignments,[]))
             M = (np.sum(E,axis=1)==0)
             if self.verbose==True:
@@ -554,7 +574,7 @@ def uniform_kernel(x,mu_,var_):
 
 
 
-@jit('float64(float64[:,:], float64[:,:], float64[:,:],int64[:,:],float64[:,:,:],float64[:,:,:],float64[:],float64[:],float64[:],int64,float64[:])',nopython=True)
+@jit('float64(float64[:,:], float64[:,:], float64[:,:],int64[:,:],float64[:,:,:],float64[:,:,:],float64[:],float64[:],float64[:],int64,float64[:])',nopython=True,cache=True)
 def hdre_loss(X,y,w_,dim_combos, pdf_combos,bin_combos, min_lim, max_lim, param_to_dim, num_dim, widths):
      var =np.copy(np.abs(w_[:num_dim])).astype(np.float64)
 
